@@ -14,6 +14,7 @@
 package promlint_test
 
 import (
+	"errors"
 	"fmt"
 	"reflect"
 	"strings"
@@ -668,7 +669,6 @@ func TestLintMetricTypeInName(t *testing.T) {
 		twoProbTest,
 		genTest("instance_memory_limit_bytes_gauge", "gauge", "gauge"),
 		genTest("request_duration_seconds_summary", "summary", "summary"),
-		genTest("request_duration_seconds_summary", "histogram", "summary"),
 		genTest("request_duration_seconds_histogram", "histogram", "histogram"),
 		genTest("request_duration_seconds_HISTOGRAM", "histogram", "histogram"),
 
@@ -734,7 +734,7 @@ request_duration_seconds{httpService="foo"} 10
 func TestLintUnitAbbreviations(t *testing.T) {
 	genTest := func(n string) test {
 		return test{
-			name: fmt.Sprintf("%s with abbreviated unit", n),
+			name: n + " with abbreviated unit",
 			in: fmt.Sprintf(`
 # HELP %s Test metric.
 # TYPE %s gauge
@@ -821,7 +821,7 @@ mc_something_total 10
 
 	prefixValidation := func(mf *dto.MetricFamily) []error {
 		if !strings.HasPrefix(mf.GetName(), "memcached_") {
-			return []error{fmt.Errorf("expected metric name to start with 'memcached_'")}
+			return []error{errors.New("expected metric name to start with 'memcached_'")}
 		}
 		return nil
 	}
@@ -840,3 +840,27 @@ mc_something_total 10
 		lintAndVerify(l2, cv)
 	})
 }
+
+func TestLintDuplicateMetric(t *testing.T) {
+	const msg = "metric not unique"
+
+	tests := []test{
+		{
+			name: "metric not unique",
+			in: `
+# HELP not_unique_total the helptext
+# TYPE not_unique_total counter
+not_unique_total{bar="abc", spam="xyz"} 1
+not_unique_total{bar="abc", spam="xyz"} 2
+`,
+			problems: []promlint.Problem{
+				{
+					Metric: "not_unique_total",
+					Text:   msg,
+				},
+			},
+		},
+	}
+
+	runTests(t, tests)
+}
